"""Module containing image transformation functions.

This module contains the function decorator
:func:`jicbioimage.core.transform.transformation` that can be used
to turn functions into image transformations.

Below is an example of how to create a transformation that inverts an image.

>>> import numpy as np
>>> @transformation
... def invert(image):
...     "Return the inverted image."
...     maximum = np.iinfo(image.dtype).max
...     maximum_array = np.ones(image.shape, dtype=image.dtype) * maximum
...     return maximum_array - image
...

"""

from functools import wraps

import numpy as np

import scipy.ndimage.filters

import skimage.io
import skimage.morphology
import skimage.exposure

from jicbioimage.core.io import AutoName, AutoWrite
from jicbioimage.core.image import Image
from jicbioimage.core.util.array import (
    normalise,
    reduce_stack,
    dtype_contract,
)

def transformation(func):
    """Function decorator to turn another function into a transformation."""
    @wraps(func)
    def func_as_transformation(*args, **kwargs):
        
        # When using transforms that return new ndarrays we lose the
        # jicbioimage.core.image.Image type and the history of the image.
        # One therefore needs to:
        #  - Extract the history from the input jicbioimage.core.image.Image.
        #  - Apply the transformation, which may return a numpy ndarray.
        #  - Force the image to the jicbioimage.core.image.Image type.
        #  - Re-attach the extracted history
        if hasattr(args[0], 'history'):
            # Working on jicbioimage.core.Image.
            history = args[0].history
        else:
            # Working on something without a history, e.g. a ndarray stack.
            history = []
        image = func(*args, **kwargs)
        image = Image.from_array(image, log_in_history=False)
        image.history = history

        image.history.append('Applied {} transform'.format(func.__name__))
        if AutoWrite.on:
            fpath = AutoName.name(func)
            im_to_save = np.copy(image)
            if AutoWrite.auto_safe_dtype:
                im_to_save = 255 * normalise(im_to_save)
                im_to_save = im_to_save.astype(np.uint8)
            try:
                skimage.io.imsave(fpath, im_to_save, 'freeimage')
            except ValueError:
                # Give a more meaningful error message.
                raise(TypeError(
                    "Cannot handle this data type: {}".format(image.dtype)))
        return image
    return func_as_transformation



"""Module for reading and writing images."""

import sys
import os
import os.path
import re
import tempfile
import subprocess
import json
from collections import namedtuple

class TemporaryFilePath(object):
    """Temporary file path context manager."""
    def __init__(self, suffix):
        self.suffix = suffix

    def __enter__(self):
        tmp_file = tempfile.NamedTemporaryFile(suffix=self.suffix, delete=False)
        self.fpath = tmp_file.name
        tmp_file.close()
        return self

    def __exit__(self, type, value, tb):
        os.unlink(self.fpath)

class AutoName(object):
    """Class for generating output file names automatically."""
    count = 0
    directory = None  #: Output directory to save images to.
    suffix = '.png'   #: Image file suffix.
    
    @classmethod
    def name(cls, func):
        """Return auto generated file name.""" 
        cls.count = cls.count + 1
        fpath = '{}_{}{}'.format(cls.count, func.__name__, cls.suffix)
        if cls.directory:
            fpath = os.path.join(cls.directory, fpath)
        return fpath
            
class AutoWrite(object):
    """Class for writing images automatically."""

    #: Whether or not auto writing of images is enabled.
    on = True

    #: Ensure image to be written has a safe dtype for writing to file.
    auto_safe_dtype = True

    
#############################################################################
# Back ends classes for storing/caching unpacked microscopy images.
#############################################################################

class FileBackend(object):
    """Class for storing image files."""

    class Entry(object):
        """Class representing a backend entry."""
        def __init__(self, base_dir, fpath):
            """Initialise a new entry; to be populated with images.

            The base name of the fpath argument is used to create a
            subdirectory in the backend directory specific for the microscopy
            file to be loaded.

            :param base_dir: backend directory
            :param fpath: path to the microscopy image of interest
            """
            fname = os.path.basename(fpath)
            self._directory = os.path.join(base_dir, fname)
            if not os.path.isdir(self.directory):
                os.mkdir(self.directory)

        @property
        def directory(self):
            """Where the images are stored."""
            return self._directory

    def __init__(self, directory):
        """Initialise a backend.

        Creates the backend directory if it does not already exist.

        :param directory: location of the backend
        """
        if not os.path.isdir(directory):
            os.mkdir(directory)
        self._directory = directory

    @property
    def directory(self):
        """Where the entries are stored."""
        return self._directory

    def new_entry(self, fpath):
        """Return a new entry; to be populated with images.

        :param fpath: path to microscopy image
        :returns: :class:`jiciimagelib.image.FileBackend.Entry` instance
        """
        return FileBackend.Entry(self.directory, fpath)


class BFConvertWrapper(object):
    """Class for unpacking microscopy files using bfconvert."""

    def __init__(self, backend):
        self.backend = backend
        self.split_order = ['s', 'c', 'z', 't']

    def split_pattern(self, win32=False):
        """Pattern used to split the input file."""
        patterns = []
        for p in self.split_order:
            if win32:
                patterns.append('_{}%%{}'.format(p.capitalize(), p))
            else:
                patterns.append('_{}%{}'.format(p.capitalize(), p))
        return ''.join(patterns)

    def _sorted_nicely(self, l):
        """Return list sorted in the way that humans expect.
        
        :param l: iterable to be sorted
        :returns: sorted list
        """
        convert = lambda text: int(text) if text.isdigit() else text
        alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
        return sorted(l, key = alphanum_key)

    def manifest(self, entry):
        """Returns manifest as a list.
        
        :param entry: :class:`jicbioimage.core.image.FileBackend.Entry`
        :returns: list
        """
        entries = []
        for fname in self._sorted_nicely(os.listdir(entry.directory)):
            if fname == 'manifest.json':
                continue
            fpath = os.path.abspath(os.path.join(entry.directory, fname))
            metadata = self.metadata_from_fname(fname)
            entries.append({"filename": fpath,
                            "series": metadata.s,
                            "channel": metadata.c,
                            "zslice": metadata.z,
                            "timepoint": metadata.t})
        return entries

    def run_command(self, input_file, output_dir=None):
        """Return the command for running bfconvert as a list.
        
        :param input_file: path to microscopy image to be converted
        :param ouput_dir: directory to write output tiff files to
        :returns: list
        """
        base_name = os.path.basename(input_file)
        name, suffix = base_name.split('.', 1)
        output_file = '{}{}.tif'.format(name, self.split_pattern())
        bfconvert = 'bfconvert'
        if sys.platform == 'win32':
            bfconvert = 'bfconvert.bat'
            output_file = '{}{}.tif'.format(name, self.split_pattern(win32=True))
        if output_dir:
            output_file = os.path.join(output_dir, output_file)
        return [bfconvert, input_file, output_file]

    def metadata_from_fname(self, fname):
        """Return meta data extracted from file name.
        
        :param fname: metadata file name
        :returns: dynamically created :class:`collections.namedtuple`
        """
        MetaData = namedtuple('MetaData', self.split_order)
        base_name = os.path.basename(fname)              # e.g. 'test_S1_C2_Z3_T4.tif'
        name, suffix = base_name.split('.')              # e.g. 'test_S1_C2_Z3_T4', 'tif'
        data = name.split('_')[-len(self.split_order):]  # e.g. ['S1', 'C2', 'Z3', 'T4']
        args = [ int(x[1:]) for x in data ]               # e.g. [1, 2, 3, 4]
        return MetaData(*args)

    def already_converted(self, fpath):
        """Return true if the file already has a manifest file in the backend.
        
        :param fpath: potential path to the manifest file
        :returns: bool
        """
        manifest_fpath = os.path.join(self.backend.directory,
                                      os.path.basename(fpath),
                                      'manifest.json')
        return os.path.isfile(manifest_fpath)
        
    def __call__(self, input_file):
        """Run the conversion.
        
        Unpacks the microscopy file and creates the manifest file.

        :param input_file: path to the microscopy file
        :returns: path to manifest file
        """
        entry = self.backend.new_entry(input_file)
        cmd = self.run_command(input_file, entry.directory)
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stderr = p.stderr.read()
        except OSError as e:
            raise(RuntimeError('bfconvert tool not found in PATH\n{}'.format(e)))
        if len(stderr) > 0:
            raise(RuntimeError(stderr))
        manifest_fpath = os.path.join(entry.directory, 'manifest.json')
        with open(manifest_fpath, 'w') as fh:
            json.dump(self.manifest(entry), fh)
        return manifest_fpath





"""Module for working with regions of interest.

This module contains the :class:`jicbioimage.core.region.Region` class,
which can be used to represent regions of interest.

One can use the
:func:`jicbioimage.core.region.Region.select_from_array` class method
to select a region from a numeric array.

>>> im = np.array([
...     [0, 0, 0, 0, 0, 0, 0],
...     [0, 1, 1, 1, 0, 0, 0],
...     [0, 1, 1, 1, 2, 2, 2],
...     [0, 1, 1, 1, 2, 2, 2],
...     [0, 0, 2, 2, 2, 2, 2],
...     [0, 0, 2, 2, 2, 2, 2],
...     [0, 0, 2, 2, 2, 2, 2]], dtype=np.uint8)
...
>>> Region.select_from_array(im, 2)
Region([[False, False, False, False, False, False, False],
       [False, False, False, False, False, False, False],
       [False, False, False, False,  True,  True,  True],
       [False, False, False, False,  True,  True,  True],
       [False, False,  True,  True,  True,  True,  True],
       [False, False,  True,  True,  True,  True,  True],
       [False, False,  True,  True,  True,  True,  True]], dtype=bool)

Alternatively a region can be created from a bitmap.

>>> import numpy as np
>>> bitmap = np.array([
...     [False, False, False, False, False, False, False],
...     [False,  True, False, False, False,  True, False],
...     [False,  True,  True, False,  True,  True, False],
...     [False,  True,  True,  True,  True,  True, False],
...     [False,  True,  True,  True,  True,  True, False],
...     [False, False,  True,  True,  True, False, False],
...     [False, False, False,  True, False, False, False]], dtype=np.bool)
...
>>> roi = Region(bitmap)

A region has several handy properties.

>>> roi.area
20
>>> roi.perimeter
14

The latter is calculated from the border.

>>> roi.border
Region([[False, False, False, False, False, False, False],
       [False,  True, False, False, False,  True, False],
       [False,  True,  True, False,  True,  True, False],
       [False,  True, False,  True, False,  True, False],
       [False,  True, False, False, False,  True, False],
       [False, False,  True, False,  True, False, False],
       [False, False, False,  True, False, False, False]], dtype=bool)

Another handy property is the convex hull.

>>> roi.convex_hull
Region([[False, False, False, False, False, False, False],
       [False,  True,  True,  True,  True,  True, False],
       [False,  True,  True,  True,  True,  True, False],
       [False,  True,  True,  True,  True,  True, False],
       [False,  True,  True,  True,  True,  True, False],
       [False, False,  True,  True,  True, False, False],
       [False, False, False,  True, False, False, False]], dtype=bool)

"""

import scipy.ndimage as nd
import numpy as np
import skimage
import skimage.morphology

class Region(np.ndarray):
    """Class representing a region of interest in an image.

    The :class:`jicbioimage.core.region.Region` class is a subclass of
    numpy.ndarray.

    However, note that it will compress any data given to it to boolean.

    >>> import numpy as np
    >>> ar = np.array([-1, 0, 1, 2])
    >>> Region(ar)
    Region([ True, False,  True,  True], dtype=bool)

    To select an particular element use the
    :func:`jicbioimage.core.region.Region.select_from_array` class method.

    >>> Region.select_from_array(ar, identifier=2)
    Region([False, False, False,  True], dtype=bool)

    """

    def __new__(cls, input_array):
        obj = np.asarray(input_array).view(cls)
        return obj.astype(bool)

    @classmethod
    def select_from_array(cls, array, identifier):
        """Return a region from a numpy array.
        
        :param array: :class:`numpy.ndarray`
        :param identifier: value representing the region to select in the array
        :returns: :class:`jicbioimage.core.region.Region`
        """

        base_array = np.zeros(array.shape)
        array_coords = np.where(array == identifier)
        base_array[array_coords] = 1

        return cls(base_array)

    @property
    def inner(self):
        """Region formed by taking non-border elements.

        :returns: :class:`jicbioimage.core.region.Region`
        """

        inner_array = nd.morphology.binary_erosion(self)
        return Region(inner_array)

    @property
    def border(self):
        """Region formed by taking border elements.

        :returns: :class:`jicbioimage.core.region.Region`
        """

        border_array = self - self.inner
        return Region(border_array)

    @property
    def convex_hull(self):
        """Region representing the convex hull.

        :returns: :class:`jicbioimage.core.region.Region`
        """
        hull_array = skimage.morphology.convex_hull_image(self)
        return Region(hull_array)

    @property
    def area(self):
        """Number of non-zero elements.

        :returns: int
        """
        return np.count_nonzero(self)

    @property
    def index_arrays(self):
        """All nonzero elements as a pair of arrays."""
        return np.where(self == True)

    @property
    def points(self):
        """Region as a list of points."""
        return list(zip(*self.index_arrays))


    @property
    def perimeter(self):
        """Return the perimiter.

        :returns: int
        """
        return self.border.area

    def dilate(self, iterations=1):
        """Return a dilated region.

        :param iterations: number of iterations to use in dilation
        :returns: :class:`jicbioimage.core.region.Region`
        """
        dilated_array = nd.morphology.binary_dilation(self, 
                                                      iterations=iterations)
        return Region(dilated_array)
